@@ -1,12 +1,12 @@
 # -*- coding: utf-8 -*-
 
-# Copyright 2018-2019 Mike Fährmann
+# Copyright 2018-2021 Mike Fährmann
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License version 2 as
 # published by the Free Software Foundation.
 
-"""Extract images from https://www.artstation.com/"""
+"""Extractors for https://www.artstation.com/"""
 
 from .common import Extractor, Message
 from .. import text, util, exception
@@ -29,7 +29,6 @@ class ArtstationExtractor(Extractor):
 
     def items(self):
         data = self.metadata()
-        yield Message.Version, 1
         yield Message.Directory, data
 
         for project in self.projects():
@@ -39,8 +38,9 @@ class ArtstationExtractor(Extractor):
 
                 if adict["has_embedded_player"] and self.external:
                     player = adict["player_embedded"]
-                    url = text.extract(player, 'src="', '"')[0]
-                    if not url.startswith(self.root):
+                    url = text.extract(player, 'src="', '"')[0] or \
+                        text.extract(player, "src='", "'")[0]
+                    if url and not url.startswith(self.root):
                         asset["extension"] = None
                         yield Message.Url, "ytdl:" + url, asset
                         continue
@@ -48,7 +48,20 @@ class ArtstationExtractor(Extractor):
                 if adict["has_image"]:
                     url = adict["image_url"]
                     text.nameext_from_url(url, asset)
-                    yield Message.Url, self._no_cache(url), asset
+
+                    url = self._no_cache(url)
+                    lhs, _, rhs = url.partition("/large/")
+                    if rhs:
+                        url = lhs + "/4k/" + rhs
+                        asset["_fallback"] = self._image_fallback(lhs, rhs)
+
+                    yield Message.Url, url, asset
+
+    @staticmethod
+    def _image_fallback(lhs, rhs):
+        yield lhs + "/large/" + rhs
+        yield lhs + "/medium/" + rhs
+        yield lhs + "/small/" + rhs
 
     def metadata(self):
         """Return general metadata"""
@@ -65,6 +78,8 @@ class ArtstationExtractor(Extractor):
         data["title"] = text.unescape(data["title"])
         data["description"] = text.unescape(text.remove_html(
             data["description"]))
+        data["date"] = text.parse_datetime(
+            data["created_at"], "%Y-%m-%dT%H:%M:%S.%f%z")
 
         assets = data["assets"]
         del data["assets"]
@@ -83,14 +98,20 @@ class ArtstationExtractor(Extractor):
         response = self.request(url, notfound="user")
         return response.json()
 
-    def _pagination(self, url, params=None):
-        if not params:
-            params = {}
+    def _pagination(self, url, params=None, json=None):
+        if json:
+            params = json
+            kwargs = {"json": json}
+        else:
+            if not params:
+                params = {}
+            kwargs = {"params": params}
+
         params["page"] = 1
         total = 0
 
         while True:
-            data = self.request(url, params=params).json()
+            data = self.request(url, **kwargs).json()
             yield from data["data"]
 
             total += len(data["data"])
@@ -122,12 +143,12 @@ class ArtstationUserExtractor(ArtstationExtractor):
     """Extractor for all projects of an artstation user"""
     subcategory = "user"
     pattern = (r"(?:https?://)?(?:(?:www\.)?artstation\.com"
-               r"/(?!artwork|projects|search)([^/?&#]+)(?:/albums/all)?"
+               r"/(?!artwork|projects|search)([^/?#]+)(?:/albums/all)?"
                r"|((?!www)\w+)\.artstation\.com(?:/projects)?)/?$")
     test = (
         ("https://www.artstation.com/gaerikim/", {
-            "pattern": r"https://\w+\.artstation\.com/p/assets"
-                       r"/images/images/\d+/\d+/\d+/large/[^/]+",
+            "pattern": r"https://\w+\.artstation\.com/p/assets/images"
+                       r"/images/\d+/\d+/\d+/(4k|large|medium|small)/[^/]+",
             "count": ">= 6",
         }),
         ("https://www.artstation.com/gaerikim/albums/all/"),
@@ -137,7 +158,8 @@ class ArtstationUserExtractor(ArtstationExtractor):
 
     def projects(self):
         url = "{}/users/{}/projects.json".format(self.root, self.user)
-        return self._pagination(url)
+        params = {"album_id": "all"}
+        return self._pagination(url, params)
 
 
 class ArtstationAlbumExtractor(ArtstationExtractor):
@@ -147,7 +169,7 @@ class ArtstationAlbumExtractor(ArtstationExtractor):
                      "{album[id]} - {album[title]}")
     archive_fmt = "a_{album[id]}_{asset[id]}"
     pattern = (r"(?:https?://)?(?:(?:www\.)?artstation\.com"
-               r"/(?!artwork|projects|search)([^/?&#]+)"
+               r"/(?!artwork|projects|search)([^/?#]+)"
                r"|((?!www)\w+)\.artstation\.com)/albums/(\d+)")
     test = (
         ("https://www.artstation.com/huimeiye/albums/770899", {
@@ -190,11 +212,11 @@ class ArtstationLikesExtractor(ArtstationExtractor):
     directory_fmt = ("{category}", "{userinfo[username]}", "Likes")
     archive_fmt = "f_{userinfo[id]}_{asset[id]}"
     pattern = (r"(?:https?://)?(?:www\.)?artstation\.com"
-               r"/(?!artwork|projects|search)([^/?&#]+)/likes/?")
+               r"/(?!artwork|projects|search)([^/?#]+)/likes/?")
     test = (
         ("https://www.artstation.com/mikf/likes", {
-            "pattern": r"https://\w+\.artstation\.com/p/assets"
-                       r"/images/images/\d+/\d+/\d+/large/[^/]+",
+            "pattern": r"https://\w+\.artstation\.com/p/assets/images"
+                       r"/images/\d+/\d+/\d+/(4k|large|medium|small)/[^/]+",
             "count": 6,
         }),
         # no likes
@@ -216,7 +238,7 @@ class ArtstationChallengeExtractor(ArtstationExtractor):
                      "{challenge[id]} - {challenge[title]}")
     archive_fmt = "c_{challenge[id]}_{asset_id}"
     pattern = (r"(?:https?://)?(?:www\.)?artstation\.com"
-               r"/contests/[^/?&#]+/challenges/(\d+)"
+               r"/contests/[^/?#]+/challenges/(\d+)"
                r"/?(?:\?sorting=([a-z]+))?")
     test = (
         ("https://www.artstation.com/contests/thu-2017/challenges/20"),
@@ -241,7 +263,6 @@ class ArtstationChallengeExtractor(ArtstationExtractor):
             self.root)
 
         challenge = self.request(challenge_url).json()
-        yield Message.Version, 1
         yield Message.Directory, {"challenge": challenge}
 
         params = {"sorting": self.sorting}
@@ -268,34 +289,38 @@ class ArtstationChallengeExtractor(ArtstationExtractor):
 class ArtstationSearchExtractor(ArtstationExtractor):
     """Extractor for artstation search results"""
     subcategory = "search"
-    directory_fmt = ("{category}", "Searches", "{search[searchterm]}")
-    archive_fmt = "s_{search[searchterm]}_{asset[id]}"
+    directory_fmt = ("{category}", "Searches", "{search[query]}")
+    archive_fmt = "s_{search[query]}_{asset[id]}"
     pattern = (r"(?:https?://)?(?:\w+\.)?artstation\.com"
                r"/search/?\?([^#]+)")
-    test = ("https://www.artstation.com/search?sorting=recent&q=ancient",)
+    test = ("https://www.artstation.com/search?q=ancient&sort_by=rank", {
+        "range": "1-20",
+        "count": 20,
+    })
 
     def __init__(self, match):
         ArtstationExtractor.__init__(self, match)
         query = text.parse_query(match.group(1))
-        self.searchterm = query.get("q", "")
-        self.order = query.get("sorting", "recent").lower()
+        self.query = query.get("q", "")
+        self.sorting = query.get("sort_by", "rank").lower()
 
     def metadata(self):
         return {"search": {
-            "searchterm": self.searchterm,
-            "order": self.order,
+            "query"  : self.query,
+            "sorting": self.sorting,
         }}
 
     def projects(self):
-        order = "likes_count" if self.order == "likes" else "published_at"
-        url = "{}/search/projects.json".format(self.root)
-        params = {
-            "direction": "desc",
-            "order": order,
-            "q": self.searchterm,
-            #  "show_pro_first": "true",
-        }
-        return self._pagination(url, params)
+        url = "{}/api/v2/search/projects.json".format(self.root)
+        return self._pagination(url, json={
+            "additional_fields": "[]",
+            "filters"          : "[]",
+            "page"             : None,
+            "per_page"         : "50",
+            "pro_first"        : "1",
+            "query"            : self.query,
+            "sorting"          : self.sorting,
+        })
 
 
 class ArtstationArtworkExtractor(ArtstationExtractor):
@@ -305,7 +330,10 @@ class ArtstationArtworkExtractor(ArtstationExtractor):
     archive_fmt = "A_{asset[id]}"
     pattern = (r"(?:https?://)?(?:\w+\.)?artstation\.com"
                r"/artwork/?\?([^#]+)")
-    test = ("https://www.artstation.com/artwork?sorting=latest",)
+    test = ("https://www.artstation.com/artwork?sorting=latest", {
+        "range": "1-20",
+        "count": 20,
+    })
 
     def __init__(self, match):
         ArtstationExtractor.__init__(self, match)
@@ -316,9 +344,7 @@ class ArtstationArtworkExtractor(ArtstationExtractor):
 
     def projects(self):
         url = "{}/projects.json".format(self.root)
-        params = self.query.copy()
-        params["page"] = 1
-        return self._pagination(url, params)
+        return self._pagination(url, self.query.copy())
 
 
 class ArtstationImageExtractor(ArtstationExtractor):
@@ -330,10 +356,10 @@ class ArtstationImageExtractor(ArtstationExtractor):
     test = (
         ("https://www.artstation.com/artwork/LQVJr", {
             "pattern": r"https?://\w+\.artstation\.com/p/assets"
-                       r"/images/images/008/760/279/large/.+",
-            "content": "1f645ce7634e44675ebde8f6b634d36db0617d3c",
+                       r"/images/images/008/760/279/4k/.+",
+            "content": "7b113871465fdc09d127adfdc2767d51cf45a7e9",
             # SHA1 hash without _no_cache()
-            # "content": "2e8aaf6400aeff2345274f45e90b6ed3f2a0d946",
+            # "content": "44b80f9af36d40efc5a2668cdd11d36d6793bae9",
         }),
         # multiple images per project
         ("https://www.artstation.com/artwork/Db3dy", {
@@ -366,3 +392,21 @@ class ArtstationImageExtractor(ArtstationExtractor):
 
     def get_project_assets(self, project_id):
         return self.assets
+
+
+class ArtstationFollowingExtractor(ArtstationExtractor):
+    """Extractor for a user's followed users"""
+    subcategory = "following"
+    pattern = (r"(?:https?://)?(?:www\.)?artstation\.com"
+               r"/(?!artwork|projects|search)([^/?#]+)/following")
+    test = ("https://www.artstation.com/gaerikim/following", {
+        "pattern": ArtstationUserExtractor.pattern,
+        "count": ">= 50",
+    })
+
+    def items(self):
+        url = "{}/users/{}/following.json".format(self.root, self.user)
+        for user in self._pagination(url):
+            url = "{}/{}".format(self.root, user["username"])
+            user["_extractor"] = ArtstationUserExtractor
+            yield Message.Queue, url, user
