@@ -28,18 +28,25 @@ class ImgbbExtractor(Extractor):
 
     def items(self):
         self.login()
-        response = self.request(self.page_url, params={"sort": self.sort})
-        if response.history and response.url.startswith(self.root):
-            raise exception.NotFoundError(self.subcategory)
+
+        url = self.page_url
+        params = {"sort": self.sort}
+        while True:
+            response = self.request(url, params=params, allow_redirects=False)
+            if response.status_code < 300:
+                break
+            url = response.headers["location"]
+            if url.startswith(self.root):
+                raise exception.NotFoundError(self.subcategory)
+
         page = response.text
         data = self.metadata(page)
         first = True
 
-        yield Message.Version, 1
         for img in self.images(page):
             image = {
                 "id"       : img["url_viewer"].rpartition("/")[2],
-                "user"     : img["user"]["username"],
+                "user"     : img["user"]["username"] if "user" in img else "",
                 "title"    : text.unescape(img["title"]),
                 "url"      : img["image"]["url"],
                 "extension": img["image"]["extension"],
@@ -79,8 +86,15 @@ class ImgbbExtractor(Extractor):
         return self.session.cookies
 
     def _pagination(self, page, endpoint, params):
-        params["page"] = 2
         data = None
+        seek, pos = text.extract(page, 'data-seek="', '"')
+        tokn, pos = text.extract(page, 'PF.obj.config.auth_token="', '"', pos)
+        params["action"] = "list"
+        params["list"] = "images"
+        params["sort"] = self.sort
+        params["seek"] = seek
+        params["page"] = 2
+        params["auth_token"] = tokn
 
         while True:
             for img in text.extract_iter(page, "data-object='", "'"):
@@ -90,6 +104,8 @@ class ImgbbExtractor(Extractor):
                     return
                 params["seek"] = data["seekEnd"]
                 params["page"] += 1
+            elif not seek or 'class="pagination-next"' not in page:
+                return
             data = self.request(endpoint, method="POST", data=params).json()
             page = data["html"]
 
@@ -98,26 +114,27 @@ class ImgbbAlbumExtractor(ImgbbExtractor):
     """Extractor for albums on imgbb.com"""
     subcategory = "album"
     directory_fmt = ("{category}", "{user}", "{album_name} {album_id}")
-    pattern = r"(?:https?://)?ibb\.co/album/([^/?&#]+)/?(?:\?([^#]+))?"
+    pattern = r"(?:https?://)?ibb\.co/album/([^/?#]+)/?(?:\?([^#]+))?"
     test = (
         ("https://ibb.co/album/i5PggF", {
             "range": "1-80",
-            "url": "570872b6eb3e11cf10b618922b780fed204c3f09",
-            "keyword": "0f2fc956728c36540c577578bd168d2459d6ae4b",
+            "url": "70afec9fcc3a6de62a6b644b487d892d8d47cf1a",
+            "keyword": "569e1d88ebdd27655387559cdf1cd526a3e1ab69",
         }),
         ("https://ibb.co/album/i5PggF?sort=title_asc", {
             "range": "1-80",
-            "url": "e2e387b8fdb3690bd75d804d0af2833112e385cd",
-            "keyword": "a307fc9d2085bdc0eb7c538c8d866c59198d460c",
+            "url": "afdf5fc95d8e09d77e8f44312f3e9b843987bb5a",
+            "keyword": "f090e14d0e5f7868595082b2c95da1309c84872d",
         }),
-        # deleted
-        ("https://ibb.co/album/fDArrF", {
-            "exception": exception.NotFoundError,
+        # no user data (#471)
+        ("https://ibb.co/album/kYKpwF", {
+            "url": "ac0abcfcb89f4df6adc2f7e4ff872f3b03ef1bc7",
+            "keyword": {"user": ""},
         }),
         # private
         ("https://ibb.co/album/hqgWrF", {
             "exception": exception.HttpError,
-        })
+        }),
     )
 
     def __init__(self, match):
@@ -133,24 +150,19 @@ class ImgbbAlbumExtractor(ImgbbExtractor):
         return {
             "album_id"  : self.album_id,
             "album_name": text.unescape(album),
-            "user"      : user.lower(),
+            "user"      : user.lower() if user else "",
         }
 
     def images(self, page):
-        seek, pos = text.extract(page, 'data-seek="', '"')
-        tokn, pos = text.extract(page, 'PF.obj.config.auth_token="', '"', pos)
+        url = text.extract(page, '"og:url" content="', '"')[0]
+        album_id = url.rpartition("/")[2].partition("?")[0]
 
         return self._pagination(page, "https://ibb.co/json", {
-            "action"    : "list",
-            "list"      : "images",
             "from"      : "album",
-            "sort"      : self.sort,
-            "albumid"   : self.album_id,
-            "seek"      : seek,
-            "auth_token": tokn,
+            "albumid"   : album_id,
             "params_hidden[list]"   : "images",
             "params_hidden[from]"   : "album",
-            "params_hidden[albumid]": self.album_id,
+            "params_hidden[albumid]": album_id,
         })
 
 
@@ -160,7 +172,7 @@ class ImgbbUserExtractor(ImgbbExtractor):
     pattern = r"(?:https?://)?([^.]+)\.imgbb\.com/?(?:\?([^#]+))?$"
     test = ("https://folkie.imgbb.com", {
         "range": "1-80",
-        "pattern": r"https?://i\.ibb\.co/\w+/[^/?&#]+",
+        "pattern": r"https?://i\.ibb\.co/\w+/[^/?#]+",
     })
 
     def __init__(self, match):
@@ -173,18 +185,10 @@ class ImgbbUserExtractor(ImgbbExtractor):
         return {"user": self.user}
 
     def images(self, page):
-        seek, pos = text.extract(page, 'data-seek="', '"')
-        tokn, pos = text.extract(page, 'PF.obj.config.auth_token="', '"', pos)
-        user, pos = text.extract(page, '.obj.resource={"id":"', '"', pos)
-
+        user = text.extract(page, '.obj.resource={"id":"', '"')[0]
         return self._pagination(page, self.page_url + "json", {
-            "action"    : "list",
-            "list"      : "images",
             "from"      : "user",
-            "sort"      : self.sort,
-            "seek"      : seek,
             "userid"    : user,
-            "auth_token": tokn,
             "params_hidden[userid]": user,
             "params_hidden[from]"  : "user",
         })
@@ -192,14 +196,14 @@ class ImgbbUserExtractor(ImgbbExtractor):
 
 class ImgbbImageExtractor(ImgbbExtractor):
     subcategory = "image"
-    pattern = r"(?:https?://)?ibb\.co/(?!album/)([^/?&#]+)"
+    pattern = r"(?:https?://)?ibb\.co/(?!album/)([^/?#]+)"
     test = ("https://ibb.co/fUqh5b", {
-        "pattern": "https://image.ibb.co/dY5FQb/Arundel-Ireeman-5.jpg",
+        "pattern": r"https://i\.ibb\.co/g3kvx80/Arundel-Ireeman-5\.jpg",
         "content": "c5a0965178a8b357acd8aa39660092918c63795e",
         "keyword": {
             "id"    : "fUqh5b",
             "title" : "Arundel Ireeman 5",
-            "url"   : "https://image.ibb.co/dY5FQb/Arundel-Ireeman-5.jpg",
+            "url"   : "https://i.ibb.co/g3kvx80/Arundel-Ireeman-5.jpg",
             "width" : 960,
             "height": 719,
             "user"  : "folkie",
@@ -225,6 +229,5 @@ class ImgbbImageExtractor(ImgbbExtractor):
         }
         image["extension"] = text.ext_from_url(image["url"])
 
-        yield Message.Version, 1
         yield Message.Directory, image
         yield Message.Url, image["url"], image
