@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-# Copyright 2017-2019 Mike Fährmann
+# Copyright 2017-2021 Mike Fährmann
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License version 2 as
@@ -9,80 +9,148 @@
 """Extractors for https://www.reddit.com/"""
 
 from .common import Extractor, Message
-from .. import text, util, extractor, exception
+from .. import text, util, exception
 from ..cache import cache
-import time
 
 
 class RedditExtractor(Extractor):
     """Base class for reddit extractors"""
     category = "reddit"
-
-    def __init__(self, match):
-        Extractor.__init__(self, match)
-        self.api = RedditAPI(self)
-        self.max_depth = int(self.config("recursion", 0))
-        self._visited = set()
+    directory_fmt = ("{category}", "{subreddit}")
+    filename_fmt = "{id}{num:? //>02} {title[:220]}.{extension}"
+    archive_fmt = "{filename}"
+    cookiedomain = None
 
     def items(self):
-        subre = RedditSubmissionExtractor.pattern
+        self.api = RedditAPI(self)
+        match_submission = RedditSubmissionExtractor.pattern.match
+        match_subreddit = RedditSubredditExtractor.pattern.match
+        match_user = RedditUserExtractor.pattern.match
+
+        parentdir = self.config("parent-directory")
+        max_depth = self.config("recursion", 0)
+        videos = self.config("videos", True)
+
         submissions = self.submissions()
+        visited = set()
         depth = 0
 
-        yield Message.Version, 1
-        with extractor.blacklist(
-                util.SPECIAL_EXTRACTORS,
-                [RedditSubredditExtractor, RedditUserExtractor]):
-            while True:
-                extra = []
-                for url, data in self._urls(submissions):
-                    if url[0] == "#":
+        while True:
+            extra = []
+
+            for submission, comments in submissions:
+                urls = []
+
+                if submission:
+                    submission["date"] = text.parse_timestamp(
+                        submission["created_utc"])
+                    yield Message.Directory, submission
+                    visited.add(submission["id"])
+                    url = submission["url"]
+                    submission["num"] = 0
+
+                    if url.startswith("https://i.redd.it/"):
+                        text.nameext_from_url(url, submission)
+                        yield Message.Url, url, submission
+
+                    elif "gallery_data" in submission:
+                        for submission["num"], url in enumerate(
+                                self._extract_gallery(submission), 1):
+                            text.nameext_from_url(url, submission)
+                            yield Message.Url, url, submission
+
+                    elif submission["is_video"]:
+                        if videos:
+                            text.nameext_from_url(url, submission)
+                            if videos == "ytdl":
+                                url = "https://www.reddit.com" + \
+                                    submission["permalink"]
+                            else:
+                                submission["_ytdl_extra"] = {
+                                    "title": submission["title"],
+                                }
+                            yield Message.Url, "ytdl:" + url, submission
+
+                    elif not submission["is_self"]:
+                        urls.append((url, submission))
+
+                elif parentdir:
+                    yield Message.Directory, comments[0]
+
+                if self.api.comments:
+                    if submission:
+                        for url in text.extract_iter(
+                                submission["selftext_html"] or "",
+                                ' href="', '"'):
+                            urls.append((url, submission))
+                    for comment in comments:
+                        for url in text.extract_iter(
+                                comment["body_html"] or "", ' href="', '"'):
+                            urls.append((url, comment))
+
+                for url, data in urls:
+                    if not url or url[0] == "#":
                         continue
                     if url[0] == "/":
                         url = "https://www.reddit.com" + url
 
-                    match = subre.match(url)
+                    match = match_submission(url)
                     if match:
                         extra.append(match.group(1))
-                    else:
+                    elif not match_user(url) and not match_subreddit(url):
                         yield Message.Queue, text.unescape(url), data
 
-                if not extra or depth == self.max_depth:
-                    return
-                depth += 1
-                submissions = (
-                    self.api.submission(sid) for sid in extra
-                    if sid not in self._visited
-                )
+            if not extra or depth == max_depth:
+                return
+            depth += 1
+            submissions = (
+                self.api.submission(sid) for sid in extra
+                if sid not in visited
+            )
 
     def submissions(self):
         """Return an iterable containing all (submission, comments) tuples"""
 
-    def _urls(self, submissions):
-        for submission, comments in submissions:
-            self._visited.add(submission["id"])
+    def _extract_gallery(self, submission):
+        gallery = submission["gallery_data"]
+        if gallery is None:
+            self.log.warning("gallery %s: deleted", submission["id"])
+            return
 
-            if not submission["is_self"]:
-                yield submission["url"], submission
+        meta = submission.get("media_metadata")
+        if meta is None:
+            self.log.warning("gallery %s: missing 'media_metadata'",
+                             submission["id"])
+            return
 
-            for url in text.extract_iter(
-                    submission["selftext_html"] or "", ' href="', '"'):
-                yield url, submission
-
-            if comments:
-                for comment in comments:
-                    for url in text.extract_iter(
-                            comment["body_html"] or "", ' href="', '"'):
-                        yield url, comment
+        for item in gallery["items"]:
+            data = meta[item["media_id"]]
+            if data["status"] != "valid" or "s" not in data:
+                self.log.warning(
+                    "gallery %s: skipping item %s ('status: %s')",
+                    submission["id"], item["media_id"], data.get("status"))
+                continue
+            src = data["s"]
+            url = src.get("u") or src.get("gif") or src.get("mp4")
+            if url:
+                yield url.partition("?")[0].replace("/preview.", "/i.", 1)
+            else:
+                self.log.error(
+                    "gallery %s: unable to fetch download URL for item %s",
+                    submission["id"], item["media_id"])
+                self.log.debug(src)
 
 
 class RedditSubredditExtractor(RedditExtractor):
     """Extractor for URLs from subreddits on reddit.com"""
     subcategory = "subreddit"
     pattern = (r"(?:https?://)?(?:\w+\.)?reddit\.com/r/"
-               r"([^/?&#]+(?:/[a-z]+)?)/?(?:\?([^#]*))?(?:$|#)")
+               r"([^/?#]+(?:/([a-z]+))?)/?(?:\?([^#]*))?(?:$|#)")
     test = (
-        ("https://www.reddit.com/r/lavaporn/"),
+        ("https://www.reddit.com/r/lavaporn/", {
+            "range": "1-20",
+            "count": ">= 20",
+        }),
         ("https://www.reddit.com/r/lavaporn/top/?sort=top&t=month"),
         ("https://old.reddit.com/r/lavaporn/"),
         ("https://np.reddit.com/r/lavaporn/"),
@@ -90,9 +158,11 @@ class RedditSubredditExtractor(RedditExtractor):
     )
 
     def __init__(self, match):
+        self.subreddit, sub, params = match.groups()
+        self.params = text.parse_query(params)
+        if sub:
+            self.subcategory += "-" + sub
         RedditExtractor.__init__(self, match)
-        self.subreddit = match.group(1)
-        self.params = text.parse_query(match.group(2))
 
     def submissions(self):
         return self.api.submissions_subreddit(self.subreddit, self.params)
@@ -102,7 +172,7 @@ class RedditUserExtractor(RedditExtractor):
     """Extractor for URLs from posts by a reddit user"""
     subcategory = "user"
     pattern = (r"(?:https?://)?(?:\w+\.)?reddit\.com/u(?:ser)?/"
-               r"([^/?&#]+(?:/[a-z]+)?)/?(?:\?([^#]*))?")
+               r"([^/?#]+(?:/([a-z]+))?)/?(?:\?([^#]*))?")
     test = (
         ("https://www.reddit.com/user/username/", {
             "count": ">= 2",
@@ -113,9 +183,11 @@ class RedditUserExtractor(RedditExtractor):
     )
 
     def __init__(self, match):
+        self.user, sub, params = match.groups()
+        self.params = text.parse_query(params)
+        if sub:
+            self.subcategory += "-" + sub
         RedditExtractor.__init__(self, match)
-        self.user = match.group(1)
-        self.params = text.parse_query(match.group(2))
 
     def submissions(self):
         return self.api.submissions_user(self.user, self.params)
@@ -125,20 +197,40 @@ class RedditSubmissionExtractor(RedditExtractor):
     """Extractor for URLs from a submission on reddit.com"""
     subcategory = "submission"
     pattern = (r"(?:https?://)?(?:"
-               r"(?:\w+\.)?reddit\.com/r/[^/?&#]+/comments|"
-               r"redd\.it"
-               r")/([a-z0-9]+)")
+               r"(?:\w+\.)?reddit\.com/(?:r/[^/?#]+/comments|gallery)"
+               r"|redd\.it)/([a-z0-9]+)")
     test = (
         ("https://www.reddit.com/r/lavaporn/comments/8cqhub/", {
+            "pattern": r"https://c2.staticflickr.com/8/7272/\w+_k.jpg",
+            "count": 1,
+        }),
+        ("https://www.reddit.com/r/lavaporn/comments/8cqhub/", {
+            "options": (("comments", 500),),
             "pattern": r"https://",
             "count": 3,
         }),
-        # ignore submission comments (#429)
-        ("https://www.reddit.com/r/lavaporn/comments/8cqhub/", {
-            "options": (("comments", 0),),
-            "pattern": r"https://c2.staticflickr.com/8/7272/\w+_k.jpg",
+        ("https://www.reddit.com/gallery/hrrh23", {
+            "url": "25b91ede15459470274dd17291424b037ed8b0ae",
+            "content": "1e7dde4ee7d5f4c4b45749abfd15b2dbfa27df3f",
+            "count": 3,
+        }),
+        # deleted gallery (#953)
+        ("https://www.reddit.com/gallery/icfgzv", {
+            "count": 0,
+        }),
+        # animated gallery items (#955)
+        ("https://www.reddit.com/r/araragi/comments/ib32hm", {
+            "pattern": r"https://i\.redd\.it/\w+\.gif",
+            "count": 2,
+        }),
+        # "failed" gallery item (#1127)
+        ("https://www.reddit.com/r/cosplay/comments/jvwaqr", {
             "count": 1,
         }),
+        # gallery with no 'media_metadata' (#2001)
+        ("https://www.reddit.com/r/kpopfap/comments/qjj04q/", {
+            "count": 0,
+        }),
         ("https://old.reddit.com/r/lavaporn/comments/2a00np/"),
         ("https://np.reddit.com/r/lavaporn/comments/2a00np/"),
         ("https://m.reddit.com/r/lavaporn/comments/2a00np/"),
@@ -159,7 +251,7 @@ class RedditImageExtractor(Extractor):
     subcategory = "image"
     archive_fmt = "{filename}"
     pattern = (r"(?:https?://)?i\.redd(?:\.it|ituploads\.com)"
-               r"/[^/?&#]+(?:\?[^#]*)?")
+               r"/[^/?#]+(?:\?[^#]*)?")
     test = (
         ("https://i.redd.it/upjtjcx2npzz.jpg", {
             "url": "0de614900feef103e580b632190458c0b62b641a",
@@ -174,34 +266,40 @@ class RedditImageExtractor(Extractor):
 
     def items(self):
         data = text.nameext_from_url(self.url)
-        yield Message.Version, 1
         yield Message.Directory, data
         yield Message.Url, self.url, data
 
 
 class RedditAPI():
-    """Minimal interface for the reddit API"""
+    """Interface for the Reddit API
+
+    Ref: https://www.reddit.com/dev/api/
+    """
     CLIENT_ID = "6N9uN0krSDE-ig"
     USER_AGENT = "Python:gallery-dl:0.8.4 (by /u/mikf1)"
 
     def __init__(self, extractor):
         self.extractor = extractor
-        self.comments = text.parse_int(extractor.config("comments", 500))
-        self.morecomments = extractor.config("morecomments", False)
-        self.refresh_token = extractor.config("refresh-token")
         self.log = extractor.log
 
-        client_id = extractor.config("client-id", self.CLIENT_ID)
-        user_agent = extractor.config("user-agent", self.USER_AGENT)
+        config = extractor.config
+        self.comments = text.parse_int(config("comments", 0))
+        self.morecomments = config("morecomments", False)
 
-        if (client_id == self.CLIENT_ID) ^ (user_agent == self.USER_AGENT):
-            self.client_id = None
-            self.log.warning(
-                "Conflicting values for 'client-id' and 'user-agent': "
-                "overwrite either both or none of them.")
+        client_id = config("client-id")
+        if client_id is None:
+            self.client_id = self.CLIENT_ID
+            self.headers = {"User-Agent": self.USER_AGENT}
         else:
             self.client_id = client_id
-            extractor.session.headers["User-Agent"] = user_agent
+            self.headers = {"User-Agent": config("user-agent")}
+
+        token = config("refresh-token")
+        if token is None or token == "cache":
+            key = "#" + self.client_id
+            self.refresh_token = _refresh_token_cache(key)
+        else:
+            self.refresh_token = token
 
     def submission(self, submission_id):
         """Fetch the (submission, comments)=-tuple for a submission id"""
@@ -209,7 +307,7 @@ class RedditAPI():
         link_id = "t3_" + submission_id if self.morecomments else None
         submission, comments = self._call(endpoint, {"limit": self.comments})
         return (submission["data"]["children"][0]["data"],
-                self._flatten(comments, link_id) if self.comments else None)
+                self._flatten(comments, link_id) if self.comments else ())
 
     def submissions_subreddit(self, subreddit, params):
         """Collect all (submission, comments)-tuples of a subreddit"""
@@ -243,13 +341,15 @@ class RedditAPI():
 
     def authenticate(self):
         """Authenticate the application by requesting an access token"""
-        access_token = self._authenticate_impl(self.refresh_token)
-        self.extractor.session.headers["Authorization"] = access_token
+        self.headers["Authorization"] = \
+            self._authenticate_impl(self.refresh_token)
 
     @cache(maxage=3600, keyarg=1)
     def _authenticate_impl(self, refresh_token=None):
         """Actual authenticate implementation"""
         url = "https://www.reddit.com/api/v1/access_token"
+        self.headers["Authorization"] = None
+
         if refresh_token:
             self.log.info("Refreshing private access token")
             data = {"grant_type": "refresh_token",
@@ -260,9 +360,9 @@ class RedditAPI():
                                    "grants/installed_client"),
                     "device_id": "DO_NOT_TRACK_THIS_DEVICE"}
 
-        auth = (self.client_id, "")
         response = self.extractor.request(
-            url, method="POST", data=data, auth=auth, fatal=False)
+            url, method="POST", headers=self.headers,
+            data=data, auth=(self.client_id, ""), fatal=False)
         data = response.json()
 
         if response.status_code != 200:
@@ -273,21 +373,28 @@ class RedditAPI():
 
     def _call(self, endpoint, params):
         url = "https://oauth.reddit.com" + endpoint
-        params["raw_json"] = 1
+        params["raw_json"] = "1"
         self.authenticate()
-        response = self.extractor.request(url, params=params, fatal=None)
+        response = self.extractor.request(
+            url, params=params, headers=self.headers, fatal=None)
+
         remaining = response.headers.get("x-ratelimit-remaining")
         if remaining and float(remaining) < 2:
-            wait = int(response.headers["x-ratelimit-reset"])
-            self.log.info("Waiting %d seconds for ratelimit reset", wait)
-            time.sleep(wait)
-        data = response.json()
+            self.extractor.wait(seconds=response.headers["x-ratelimit-reset"])
+            return self._call(endpoint, params)
+
+        try:
+            data = response.json()
+        except ValueError:
+            raise exception.StopExtraction(text.remove_html(response.text))
+
         if "error" in data:
             if data["error"] == 403:
                 raise exception.AuthorizationError()
             if data["error"] == 404:
                 raise exception.NotFoundError()
-            raise Exception(data["message"])
+            self.log.debug(data)
+            raise exception.StopExtraction(data.get("message"))
         return data
 
     def _pagination(self, endpoint, params):
@@ -298,17 +405,24 @@ class RedditAPI():
         while True:
             data = self._call(endpoint, params)["data"]
 
-            for submission in data["children"]:
-                submission = submission["data"]
-                if (date_min <= submission["created_utc"] <= date_max and
-                        id_min <= self._decode(submission["id"]) <= id_max):
-                    if submission["num_comments"] and self.comments:
-                        try:
-                            yield self.submission(submission["id"])
-                        except exception.AuthorizationError:
-                            pass
-                    else:
-                        yield submission, None
+            for child in data["children"]:
+                kind = child["kind"]
+                post = child["data"]
+
+                if (date_min <= post["created_utc"] <= date_max and
+                        id_min <= self._decode(post["id"]) <= id_max):
+
+                    if kind == "t3":
+                        if post["num_comments"] and self.comments:
+                            try:
+                                yield self.submission(post["id"])
+                            except exception.AuthorizationError:
+                                pass
+                        else:
+                            yield post, ()
+
+                    elif kind == "t1" and self.comments:
+                        yield None, (post,)
 
             if not data["after"]:
                 return
@@ -337,3 +451,10 @@ class RedditAPI():
     @staticmethod
     def _decode(sid):
         return util.bdecode(sid, "0123456789abcdefghijklmnopqrstuvwxyz")
+
+
+@cache(maxage=100*365*24*3600, keyarg=0)
+def _refresh_token_cache(token):
+    if token and token[0] == "#":
+        return None
+    return token
