@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 
-# Copyright 2015-2019 Mike Fährmann
+# Copyright 2015-2021 Mike Fährmann
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License version 2 as
@@ -9,27 +9,21 @@
 
 import os
 import sys
+import unittest
+
 import re
 import json
 import hashlib
-import unittest
-from gallery_dl import extractor, util, job, config, exception
+import datetime
 
+sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
+from gallery_dl import \
+    extractor, util, job, config, exception, formatter  # noqa E402
 
-# these don't work on Travis CI
-TRAVIS_SKIP = {
-    "exhentai", "kissmanga", "mangafox", "dynastyscans", "nijie", "bobx",
-    "archivedmoe", "archiveofsins", "thebarchive", "fireden", "4plebs",
-    "sankaku", "idolcomplex", "mangahere", "readcomiconline", "mangadex",
-    "sankakucomplex", "warosu", "fuskator",
-}
 
 # temporary issues, etc.
 BROKEN = {
-    "hentaifox",
-    "livedoor",
-    "mangapark",
-    "yaplog",
+    "photobucket",
 }
 
 
@@ -56,10 +50,11 @@ class TestExtractorResults(unittest.TestCase):
         if result:
             if "options" in result:
                 for key, value in result["options"]:
-                    config.set(key.split("."), value)
+                    key = key.split(".")
+                    config.set(key[:-1], key[-1], value)
             if "range" in result:
-                config.set(("image-range",), result["range"])
-                config.set(("chapter-range",), result["range"])
+                config.set((), "image-range"  , result["range"])
+                config.set((), "chapter-range", result["range"])
             content = "content" in result
         else:
             content = False
@@ -79,20 +74,26 @@ class TestExtractorResults(unittest.TestCase):
             pass
         except exception.HttpError as exc:
             exc = str(exc)
-            if re.match(r"5\d\d: ", exc) or \
+            if re.match(r"'5\d\d ", exc) or \
                     re.search(r"\bRead timed out\b", exc):
                 self._skipped.append((url, exc))
                 self.skipTest(exc)
             raise
 
-        # test archive-id uniqueness
-        self.assertEqual(len(set(tjob.archive_list)), len(tjob.archive_list))
+        if result.get("archive", True):
+            self.assertEqual(
+                len(set(tjob.archive_list)),
+                len(tjob.archive_list),
+                "archive-id uniqueness",
+            )
 
         if tjob.queue:
             # test '_extractor' entries
             for url, kwdict in zip(tjob.url_list, tjob.kwdict_list):
                 if "_extractor" in kwdict:
                     extr = kwdict["_extractor"].from_url(url)
+                    if extr is None and not result.get("extractor", True):
+                        continue
                     self.assertIsInstance(extr, kwdict["_extractor"])
                     self.assertEqual(extr.url, url)
         else:
@@ -105,7 +106,12 @@ class TestExtractorResults(unittest.TestCase):
             self.assertEqual(result["url"], tjob.url_hash.hexdigest())
 
         if "content" in result:
-            self.assertEqual(result["content"], tjob.content_hash.hexdigest())
+            expected = result["content"]
+            digest = tjob.content_hash.hexdigest()
+            if isinstance(expected, str):
+                self.assertEqual(digest, expected, "content")
+            else:  # assume iterable
+                self.assertIn(digest, expected, "content")
 
         if "keyword" in result:
             expected = result["keyword"]
@@ -142,9 +148,20 @@ class TestExtractorResults(unittest.TestCase):
                 self._test_kwdict(value, test)
             elif isinstance(test, type):
                 self.assertIsInstance(value, test, msg=key)
+            elif isinstance(test, list):
+                subtest = False
+                for idx, item in enumerate(test):
+                    if isinstance(item, dict):
+                        subtest = True
+                        self._test_kwdict(value[idx], item)
+                if not subtest:
+                    self.assertEqual(value, test, msg=key)
             elif isinstance(test, str):
                 if test.startswith("re:"):
                     self.assertRegex(value, test[3:], msg=key)
+                elif test.startswith("dt:"):
+                    self.assertIsInstance(value, datetime.datetime, msg=key)
+                    self.assertEqual(str(value), test[3:], msg=key)
                 elif test.startswith("type:"):
                     self.assertEqual(type(value).__name__, test[5:], msg=key)
                 else:
@@ -185,13 +202,16 @@ class ResultJob(job.DownloadJob):
         self._update_url(url)
         self._update_kwdict(kwdict)
         self._update_archive(kwdict)
-        self._update_content(url)
+        self._update_content(url, kwdict)
         self.format_filename(kwdict)
 
     def handle_directory(self, kwdict):
         self._update_kwdict(kwdict, False)
         self.format_directory(kwdict)
 
+    def handle_metadata(self, kwdict):
+        pass
+
     def handle_queue(self, url, kwdict):
         self.queue = True
         self._update_url(url)
@@ -204,7 +224,7 @@ class ResultJob(job.DownloadJob):
     def _update_kwdict(self, kwdict, to_list=True):
         if to_list:
             self.kwdict_list.append(kwdict.copy())
-        kwdict = self._filter(kwdict)
+        kwdict = util.filter_dict(kwdict)
         self.kwdict_hash.update(
             json.dumps(kwdict, sort_keys=True, default=str).encode())
 
@@ -213,9 +233,10 @@ class ResultJob(job.DownloadJob):
         self.archive_list.append(archive_id)
         self.archive_hash.update(archive_id.encode())
 
-    def _update_content(self, url):
+    def _update_content(self, url, kwdict):
         if self.content:
             scheme = url.partition(":")[0]
+            self.fileobj.kwdict = kwdict
             self.get_downloader(scheme).download(url, self.fileobj)
 
 
@@ -250,14 +271,14 @@ class TestPathfmt():
         return 0
 
 
-class TestFormatter(util.Formatter):
+class TestFormatter(formatter.StringFormatter):
 
     @staticmethod
     def _noop(_):
         return ""
 
     def _apply_simple(self, key, fmt):
-        if key == "extension" or "._format_optional." in repr(fmt):
+        if key == "extension" or "_parse_optional." in repr(fmt):
             return self._noop
 
         def wrap(obj):
@@ -265,7 +286,7 @@ class TestFormatter(util.Formatter):
         return wrap
 
     def _apply(self, key, funcs, fmt):
-        if key == "extension" or "._format_optional." in repr(fmt):
+        if key == "extension" or "_parse_optional." in repr(fmt):
             return self._noop
 
         def wrap(obj):
@@ -279,36 +300,45 @@ class TestFormatter(util.Formatter):
 def setup_test_config():
     name = "gallerydl"
     email = "gallerydl@openaliasbox.org"
+    email2 = "gallerydl@protonmail.com"
 
     config.clear()
-    config.set(("cache", "file"), ":memory:")
-    config.set(("downloader", "part"), False)
-    config.set(("downloader", "adjust-extensions"), False)
-    config.set(("extractor", "timeout"), 60)
-    config.set(("extractor", "username"), name)
-    config.set(("extractor", "password"), name)
-    config.set(("extractor", "nijie"    , "username"), email)
-    config.set(("extractor", "seiga"    , "username"), email)
+    config.set(("cache",), "file", None)
+    config.set(("downloader",), "part", False)
+    config.set(("downloader",), "adjust-extensions", False)
+    config.set(("extractor" ,), "timeout" , 60)
+    config.set(("extractor" ,), "username", name)
+    config.set(("extractor" ,), "password", name)
 
-    config.set(("extractor", "danbooru" , "username"), None)
-    config.set(("extractor", "instagram", "username"), None)
-    config.set(("extractor", "imgur"    , "username"), None)
-    config.set(("extractor", "twitter"  , "username"), None)
+    config.set(("extractor", "nijie")     , "username", email)
+    config.set(("extractor", "seiga")     , "username", email)
+    config.set(("extractor", "pinterest") , "username", email2)
+    config.set(("extractor", "pinterest") , "username", None)  # login broken
 
-    config.set(("extractor", "mangoxo"  , "username"), "LiQiang3")
-    config.set(("extractor", "mangoxo"  , "password"), "5zbQF10_5u25259Ma")
+    config.set(("extractor", "newgrounds"), "username", "d1618111")
+    config.set(("extractor", "newgrounds"), "password", "d1618111")
 
-    config.set(("extractor", "deviantart", "client-id"), "7777")
-    config.set(("extractor", "deviantart", "client-secret"),
+    config.set(("extractor", "mangoxo")   , "username", "LiQiang3")
+    config.set(("extractor", "mangoxo")   , "password", "5zbQF10_5u25259Ma")
+
+    for category in ("danbooru", "instagram", "twitter", "subscribestar",
+                     "e621", "inkbunny", "tapas", "pillowfort", "mangadex"):
+        config.set(("extractor", category), "username", None)
+
+    config.set(("extractor", "mastodon.social"), "access-token",
+               "Blf9gVqG7GytDTfVMiyYQjwVMQaNACgf3Ds3IxxVDUQ")
+
+    config.set(("extractor", "deviantart"), "client-id", "7777")
+    config.set(("extractor", "deviantart"), "client-secret",
                "ff14994c744d9208e5caeec7aab4a026")
 
-    config.set(("extractor", "tumblr", "api-key"),
+    config.set(("extractor", "tumblr"), "api-key",
                "0cXoHfIqVzMQcc3HESZSNsVlulGxEXGDTTZCDrRrjaa0jmuTc6")
-    config.set(("extractor", "tumblr", "api-secret"),
+    config.set(("extractor", "tumblr"), "api-secret",
                "6wxAK2HwrXdedn7VIoZWxGqVhZ8JdYKDLjiQjL46MLqGuEtyVj")
-    config.set(("extractor", "tumblr", "access-token"),
+    config.set(("extractor", "tumblr"), "access-token",
                "N613fPV6tOZQnyn0ERTuoEZn0mEqG8m2K8M3ClSJdEHZJuqFdG")
-    config.set(("extractor", "tumblr", "access-token-secret"),
+    config.set(("extractor", "tumblr"), "access-token-secret",
                "sgOA7ZTT4FBXdOGGVV331sSp0jHYp4yMDRslbhaQf7CaS71i4O")
 
 
@@ -333,8 +363,6 @@ def generate_tests():
         del sys.argv[1:]
     else:
         skip = set(BROKEN)
-        if "CI" in os.environ and "TRAVIS" in os.environ:
-            skip |= set(TRAVIS_SKIP)
         if skip:
             print("skipping:", ", ".join(skip))
         fltr = lambda c, bc: c not in skip  # noqa: E731
@@ -342,7 +370,7 @@ def generate_tests():
     # filter available extractor classes
     extractors = [
         extr for extr in extractor.extractors()
-        if fltr(extr.category, getattr(extr, "basecategory", None))
+        if fltr(extr.category, extr.basecategory)
     ]
 
     # add 'test_...' methods
